import logging

from typing import Tuple, Any

import tornado.tcpserver
import tornado.iostream

import gb.protocol
import gb.mode


log = logging.getLogger(__name__)


class GopherServer(tornado.tcpserver.TCPServer):
    """Generic tornado TCPServer which speaks the Gopher protocol. The Gopher
    protocol is defined in RFC 1436:

    https://tools.ietf.org/html/rfc1436."""

    # XXX TODO has to come from somewhere!
    encoding: str
    mode: gb.mode.Mode

    async def handle_stream(
        self, stream: tornado.iostream.IOStream, address: Tuple[Any, ...]
    ) -> None:
        """A new incoming connection. We wait silently until the client sends
        a selector through after which we clean and parse that selector."""

        log.debug("Accepted connection from %s", address)

        while True:
            try:
                # clean this up, split on \n but clean both
                selector_raw = await stream.read_until(
                    gb.protocol.crlf.encode(self.encoding),
                )
                selector_dec = selector_raw.decode(self.encoding)

                # clean this up, split on \n but clean both with \r
                selector_dec = gb.protocol.clean_selector(selector_dec)

                # If this is not a valid selector we immediately terminate
                # the connection
                if not gb.protocol.is_valid_selector(selector_dec):
                    log.warn(
                        "%s requested invalid selector %r, terminating.",
                        address,
                        selector_dec,
                    )
                    await self.close_stream(stream)
                    break

                log.info("%s requested %r", address, selector_dec)

                # Use our mode specific lookup to get our response
                resp_raw = await self.lookup(stream, selector_dec)
                resp_enc = resp_raw.encode(self.encoding)

                # Write and exit the connection
                await stream.write(resp_enc)
                await self.close_stream(stream)
            except ValueError:
                log.warning("Looking up selector %r failed", selector_dec)
                await self.close_stream(stream)
                break
            except tornado.iostream.StreamClosedError:
                log.debug("Lost connection from %s", address)
                break

    async def close_stream(self, stream: tornado.iostream.IOStream) -> None:
        """Gopher connections are closed by writing a . on a single line then
        closing the underlying transport."""
        await stream.write(gb.protocol.eof.encode(self.encoding))
        await stream.write(gb.protocol.crlf.encode(self.encoding))
        stream.close()

    async def lookup(self, stream: tornado.iostream.IOStream, data: str) -> str:
        """Lookup a selector on our mode."""
        return self.mode.lookup(data)


class ImplicitGopherServer(GopherServer):
    """The implicit gopher server serves files from a given path recursively
    while auto generating indexes for directories. If magic is enabled then
    the mode will auto-guess filetypes."""

    def __init__(
        self, path: str, host: str, port: int, magic: bool, encoding: str
    ) -> None:
        super().__init__()

        log.info("Starting ImplicitGopherServer with path %s", path)

        self.encoding = encoding
        self.mode = gb.mode.ImplicitMode(path, host, port, magic)



import os

import gb.protocol
import gb.document
import gb.entry
import gb.safe
import gb.magic


class Mode:
    def __init__(
        self, base_path: str, host: str, port: int, magic: bool
    ) -> None:
        self.base_path = os.path.abspath(base_path)
        self.host = host
        self.port = port

        self.magic = magic

    def lookup(self, path: str) -> str:
        raise NotImplementedError()


class ImplicitMode(Mode):
    """ImplicitMode looks up files recursively within a given path auto-
    generating any required indexes."""

    def lookup(self, path: str) -> str:
        """Look up a path within our base path and return the contents in
        pre-rendered Gopher format!"""

        path = gb.safe.relativize(self.base_path, path)

        if os.path.islink(path):
            raise ValueError()

        if os.path.isdir(path):
            return self._directory(path)
        elif os.path.isfile(path):
            return self._file(path)
        else:
            raise ValueError()

    def _directory(self, path: str) -> str:
        """Render the files in a directory as gopher data."""

        response = gb.document.Document()

        for entry in os.listdir(path):
            entry = os.path.join(path, entry)

            if os.path.isdir(entry):
                item = gb.entry.Directory
            elif os.path.isfile(entry):
                if self.magic:
                    item = gb.magic.guess_type(entry)  # type: ignore
                else:
                    item = gb.entry.Binary  # type: ignore

            response.add_entry(
                item(
                    self.host,
                    self.port,
                    os.path.basename(entry),
                    entry[len(self.base_path) :],
                )
            )

        return str(response)

    def _file(self, path: str) -> str:
        # XXX this fails when in utf8 (or in general), read bytes and then
        # XXX either use surrogateescape (or just return bytes) or only do
        # XXX so for non-readable files
        with open(path) as f:
            return f.read()



import sys
import ipaddress
import logging

import click
import tornado.iostream

import gbs.server


log = logging.getLogger(__name__)

modes = {
    "implicit": gbs.server.ImplicitGopherServer,
}


def bail(message: str) -> None:
    """Write a message to stderr and exit unsuccesfully."""
    print(message, file=sys.stderr)
    raise SystemExit(1)


@click.command()
@click.option(
    "--mode",
    "-m",
    required=True,
    type=click.Choice([str(mode) for mode in modes.keys()]),
    help="Mode to run as.",
)
@click.option(
    "--host",
    "-h",
    default="127.0.0.1",
    show_default=True,
    help="IP address to listen on.",
)
@click.option(
    "--port", "-p", default=7070, show_default=True, help="Port to listen on."
)
@click.option(
    "--verbose",
    "-v",
    count=True,
    help="Verbosity, passing more heightens the verbosity.",
)
@click.option(
    "--magic",
    default=False,
    help="Enable magic mode to try and guess filetypes by extension and content.",
    show_default=True,
    is_flag=True,
)
@click.option(
    "--utf8",
    "-u",
    default=False,
    help="Use UTF-8 as the default encoding instead of the standard ISO-8859-1",
    is_flag=True,
)
@click.argument(
    "path", required=True, type=click.Path(exists=True), envvar="GB_PATH"
)
def main(
    mode: str,
    host: str,
    port: int,
    path: click.Path,
    verbose: int,
    magic: bool,
    utf8: bool,
) -> int:
    """`gbs` or gopherball is a modern server for the Gopher protocol."""

    if port < 0 or port > 65535:
        bail("Invalid port supplied.")

    try:
        host = str(ipaddress.ip_address(host))
    except ValueError:
        bail("Unparseable IP supplied.")

    if mode not in modes:
        bail("Invalid mode supplied.")

    if utf8:
        encoding = "utf8"
    else:
        encoding = "iso-8859-1"

    logging.basicConfig(level=logging.INFO)

    server = modes[mode](str(path), "localhost", 7070, magic, encoding)
    server.listen(port)
    server.start(0)

    tornado.ioloop.IOLoop.current().start()

    return 0


if __name__ == "__main__":
    raise SystemExit(main(auto_envvar_prefix="GB"))
